/*
 * Disk.c
 *
 *  Created on: Oct 18, 2009
 *      Author: chwolfe2
 */


#include "string.h"
#include "Disk.h"
#include "HostInterface.h"
#include "FileSystem.h"
#include "Globals.h"
#include <stdio.h>
#include <assert.h>


// define some constants to aid in device/partition type recognition
const char* SysBlockDevs[]={"sda"};
const int NumSysBlockDevs = 1;
//const char* SysPartitions[]={"sda1","sda2","sda5"};
const char* SysPartitions[]={"sda1","sda2"};
//const int NumSysPartitions = 3;
const int NumSysPartitions = 2;
const char* IntMultiDiskPartitions[]={"sdb", "sdc", "sdd", "sde", "sdf"};
const int NumIntMultiDiskPartitions = 5;
const char* IntMultiDiskDevs[]={"md0", "md1", "md2", "md3", "md4", "md5"};
const int NumIntMultiDiskDevs = 6;
const char * DriveTypeString[]={
	"DRIVETYPE_UNKNOWN",
	"DRIVETYPE_SYSTEM_BLOCK_DEVICE",
	"DRIVETYPE_SYSTEM_PARTITION",
	"DRIVETYPE_SYSTEM_OPTICAL",
	"DRIVETYPE_SYSTEM_FLOPPY",
	"DRIVETYPE_INTERNAL_BLOCK_DEVICE",
	"DRIVETYPE_INTERNAL_MULTIDISK_DEVICE",
	"DRIVETYPE_EXTERNAL_BLOCK_DEVICE",
	"DRIVETYPE_EXTERNAL_PARTITION"
};



// record to hold information regarding detected devices
Disk 		devices[MAX_NUM_DEVICES_TO_ENUMERATE];
size_t 		devicesCount=0;
boolean 	deviceRecordLocked=0;

int DiskGetType(int index){
	if (index<0 || index>devicesCount) return DRIVETYPE_UNKNOWN;
	return devices[index].type;
}
size_t DiskGetFreeSpace(int index){
	if (index<0 || index>devicesCount) return 0;
	switch (DiskGetType(index)){
		case DRIVETYPE_INTERNAL_MULTIDISK_DEVICE:
			return FileSystem_GetFreeSpace(globals.fileSystem);
			break;
		case DRIVETYPE_EXTERNAL_PARTITION:
			return devices[index].bytesTotal - devices[index].bytesUsed;
			break;
		default:
			return 0;
	}

}
int DiskFindIndexByName(char * deviceName){
	int i;
	for (i=0;i<devicesCount;i++){
		if (strcmp(deviceName,devices[i].deviceName)==0)	return i;
		//printf("%s is not %s\n",deviceName,devices[i].deviceName);
	}
	return -1;
}


boolean matchStringInStringList(char* string, const char** matchList, int matchListSize ){
	int i;
	for (i=0;i<matchListSize;i++){
		if (strcmp(string,matchList[i])==0)
			return true;
	}
	return false;
}
StatusCode DiskLockDeviceInfo(){
	if (deviceRecordLocked) return FAILURE;
	deviceRecordLocked=true;
	return SUCCESS;
}
StatusCode DiskUnlockDeviceInfo(){
	deviceRecordLocked=false;
	return SUCCESS;
}
StatusCode DiskLockDevice(char * deviceName){
	int i=DiskFindIndexByName(deviceName);
	if (i<0) return FAILURE;
	if (devices[i].locked) return FAILURE;
	devices[i].locked=true;
	return SUCCESS;
}
StatusCode DiskUnlockDevice(char * deviceName){
	int i=DiskFindIndexByName(deviceName);
	if (i<0) return FAILURE;
	devices[i].locked=false;
	return SUCCESS;
}
char * getDevShortName(char*deviceName){
	return &deviceName[5];
}
StatusCode Disk_IdentifyAll(){
	if (deviceRecordLocked) return FAILURE;
	char buffer[2048];
	int i=0;
	int j=0;
	if (issueShellCommand("cat /proc/partitions  | awk '{if (NR>2)print $4}'", buffer, 2048) == -1) {
		return FAILURE;
	}
	if (strlen(buffer)==0){
		return FAILURE;
	}
	// now process the output to determine which devices of what type are present
	devicesCount=0;
	char * token =strtok(buffer," \n");
	char cmdTemp [1024];
	// pass 1 finds instantly recognizable drive device IDs and assigns type/size information
	while (token!=NULL){
		sprintf(devices[devicesCount].deviceName,"/dev/%s",token);
		sprintf(cmdTemp,"fdisk -l \"%s\" 2>/dev/null | head -n 2 | tail -n 1 | awk '{print $5}'",token);
		ShellGetNumber(cmdTemp,&devices[devicesCount].bytesTotal);
		if (matchStringInStringList(token, SysBlockDevs, NumSysBlockDevs)) {
			devices[devicesCount].type = DRIVETYPE_SYSTEM_BLOCK_DEVICE;
			devices[devicesCount].mounted=false;
			devices[devicesCount].locked=true;
		} else if(matchStringInStringList(token, SysPartitions, NumSysPartitions)) {
			devices[devicesCount].type = DRIVETYPE_SYSTEM_PARTITION;
			devices[devicesCount].mounted=true;
		} else if(matchStringInStringList(token, IntMultiDiskPartitions, NumIntMultiDiskPartitions)) {
			devices[devicesCount].type = DRIVETYPE_INTERNAL_BLOCK_DEVICE;
			devices[devicesCount].mounted=false;
			devices[devicesCount].locked=true;
		} else if(matchStringInStringList(token, IntMultiDiskDevs, NumIntMultiDiskDevs)) {
			devices[devicesCount].type = DRIVETYPE_INTERNAL_MULTIDISK_DEVICE;
		} else if(strncmp(token, "fd", 2)==0) {
			devices[devicesCount].type = DRIVETYPE_SYSTEM_FLOPPY;
		} else if(strncmp(token, "sr", 2)==0) {
			devices[devicesCount].type = DRIVETYPE_SYSTEM_OPTICAL;
		} else {
			// at this point, options are DRIVETYPE_UNKNOWN, DRIVETYPE_INTERNAL_BLOCK_DEVICE, DRIVETYPE_EXTERNAL_BLOCK_DEVICE, DRIVETYPE_EXTERNAL_PARTITION
			// TODO: advanced type determination
			devices[devicesCount].type = DRIVETYPE_UNKNOWN;

		}
		devicesCount++;
		token = strtok(NULL," \n"); //continue tokenization
	}
	/*
	// TODO: occupied size determination
			sprintf(cmdTemp,"",token);
			ShellGetNumber(cmdTemp,&devices[devicesCount].bytesTotal);
			devices[devicesCount].bytesUsed=999999999999l;*/
	// pass 2 looks at multi-disk arrays and assigns the proper type to array member devices
	for (i=0;i<devicesCount; i++){
		if (devices[i].type == DRIVETYPE_INTERNAL_MULTIDISK_DEVICE){
			sprintf(cmdTemp,
					"cat /proc/mdstat | awk '/^%s/,/chunks$/ {for(i=4;i<NF;i++){print $(i+1)}}' | sed 's/\[[0-9]*\]//'"
					,getDevShortName(devices[i].deviceName));
			//printf("\n>>>>>>> %s\n",cmdTemp);
			if (issueShellCommand(cmdTemp, buffer, 2048) == -1) {
				return FAILURE;
			} else {
				// now buffer contains a list of devices belonging to our multi-disk device
				token =strtok(buffer," \n");
				while (token!=NULL){
					for(j=0;j<devicesCount; j++){
						if (strcmp(token,devices[j].deviceName)==0){
							devices[j].type = DRIVETYPE_INTERNAL_BLOCK_DEVICE;
						}
					}
					token = strtok(NULL," \n"); //continue tokenization
				}
			}
		}
	}
	// pass 3 looks for all remaining unknown partitions and assumes them to be external devices
	for (i=0;i<devicesCount; i++){
		if (devices[i].type == DRIVETYPE_UNKNOWN){
			if (strcspn (devices[i].deviceName,"1234567890") == strlen(devices[i].deviceName))
				devices[i].type = DRIVETYPE_EXTERNAL_BLOCK_DEVICE;
			else
				devices[i].type = DRIVETYPE_EXTERNAL_PARTITION;
		}
		printf("Disk_IdentifyAll(): Device found: '%s' of type '%s'\n", devices[i].deviceName, DriveTypeString[devices[i].type]);
	}
	return (devicesCount) ? SUCCESS : FAILURE;
}
StatusCode Disk_GetDiskInfo(char *  deviceName, Disk* disk){
	int i = DiskFindIndexByName(deviceName);
	if (i!=-1){
		*disk=devices[i];
		return SUCCESS;
	}
	// if we got here, then we didn't find the drive
	return FAILURE;

}
StatusCode Disk_Format(char* deviceName){
	int i = DiskFindIndexByName(deviceName);
	char buffer[2048];
	if (i!=-1){
		if (devices[i].mounted || devices[i].locked) return FAILURE;
		if ((devices[i].type == DRIVETYPE_EXTERNAL_PARTITION) || ((devices[i].type == DRIVETYPE_INTERNAL_MULTIDISK_DEVICE))){
			if (devices[i].type == DRIVETYPE_EXTERNAL_PARTITION){
				if (issueShellCommand("mke2fs %s -q", buffer, 2048) != -1) return SUCCESS;
			} else {
				return FileSystem_Create(&devices[i]);
			}
		}
	}
	// if we got here, then we didn't format the drive, or an error occurred
	return FAILURE;
}

StatusCode Disk_Mount(char* deviceName){
	int i = DiskFindIndexByName(deviceName);
	char buffer[2048];
	char cmdTemp [1024];
	//printf("mount()\n");
	if (i!=-1){
		if (devices[i].mounted || devices[i].locked) return FAILURE;
		//printf("not mounted locked\n");
		if ((devices[i].type == DRIVETYPE_EXTERNAL_PARTITION) || ((devices[i].type == DRIVETYPE_INTERNAL_MULTIDISK_DEVICE))){
			//printf("good type\n");
			if (devices[i].type == DRIVETYPE_EXTERNAL_PARTITION){
				sprintf(cmdTemp,"mkdir -p /LWA_EXT/%s",deviceName);
				if (issueShellCommand(cmdTemp, buffer, 2048) == -1) return FAILURE;
				//printf("mkdir good\n");
				sprintf(cmdTemp,"mount %s -t ext2 /LWA_EXT/%s",deviceName,deviceName);
				if (issueShellCommand(cmdTemp, buffer, 2048) != -1) return SUCCESS;
				//printf("mount not good\n");
			} else {
				devices[i].mounted=true;
				return SUCCESS;
			}
		}
	}
	// if we got here, then we didn't mount the drive, or an error occurred
	return FAILURE;
}
StatusCode Disk_Unmount(char* deviceName){
	int i = DiskFindIndexByName(deviceName);
	char buffer[2048];
	char cmdTemp [1024];
	if (i!=-1){
		if (!devices[i].mounted || devices[i].locked) return FAILURE;
		if ((devices[i].type == DRIVETYPE_EXTERNAL_PARTITION) || ((devices[i].type == DRIVETYPE_INTERNAL_MULTIDISK_DEVICE))){
			if (devices[i].type == DRIVETYPE_EXTERNAL_PARTITION){
				sprintf(cmdTemp,"umount %s",deviceName);
				if (issueShellCommand(cmdTemp, buffer, 2048) != -1) return SUCCESS;
			} else {
				devices[i].mounted=true;
				return SUCCESS;
			}
		}
	}
	// if we got here, then we didn't unmount the drive, or an error occurred
	return FAILURE;
}

size_t Disk_GetArrayNumDisks(Disk* disk){
	assert(disk!=NULL);
	assert(disk->type==DRIVETYPE_INTERNAL_MULTIDISK_DEVICE);
	char cmdTemp[1024];
	size_t disks=0;
	sprintf(cmdTemp,"mdadm --detail %s | grep \"Raid Devices\" | awk '{print $4}'",disk->deviceName);
	ShellGetNumber(cmdTemp,&disks);
	if (ShellGetNumber(cmdTemp,&disks)==FAILURE) {
		printf("Error: couldn't get RAID array number of disks.\n");
		return 0l;
	}
	return disks;
}

size_t Disk_GetArraySize(Disk* disk){
	assert(disk!=NULL);
	assert(disk->type==DRIVETYPE_INTERNAL_MULTIDISK_DEVICE);
	char cmdTemp[1024];
	size_t size=0;
	sprintf(cmdTemp,"mdadm --detail %s | grep \"Array Size\" | awk '{print $4}'",disk->deviceName);
	if (ShellGetNumber(cmdTemp,&size)==FAILURE) {
		printf("Error: couldn't get RAID array size.\n");
		return 0l;
	}
	return size*1024;
}

size_t Disk_GetArrayChunkSize(Disk* disk){
	assert(disk!=NULL);
	assert(disk->type==DRIVETYPE_INTERNAL_MULTIDISK_DEVICE);
	char cmdTemp[1024];
	size_t size=0;

	sprintf(cmdTemp,"mdadm --detail %s | grep \"Chunk Size\" | awk '{print $4}' | sed 's/K/\\*1024/' | bc",disk->deviceName);
	if (ShellGetNumber(cmdTemp,&size)==FAILURE) {
		printf("Error: couldn't get RAID array size.\n");
		return 0l;
	}
	return size;
}
