/*
 * FileSystem.h
 *
 *  Created on: Oct 25, 2009
 *      Author: chwolfe2
 */
#ifndef FILESYSTEM_H_
#define FILESYSTEM_H_
#include "Defines.h"
#include "Disk.h"
#include "Time.h"
#include <aio.h>


typedef int FILEMODE;

#define READ 							0
#define WRITE 							1
#define FILE_HEADER_SIZE 				4096
#define FILE_HEADER_RESERVED_SPACE      2944
#define FILE_MAX_NAME_LENGTH 		    1023
#define MAX_NUMBER_FILES 			    1023
#define DEFAULT_VOLUME_NAME				"MCS-DRSU-Storage"



typedef struct __FileInfo{
	union {
		struct {
			size_t startPosition;
			size_t stopPosition;
			size_t size;
			size_t isOpen;
			size_t currentPosition;
			size_t tags;
			FILEMODE mode;
			TimeStamp creationDate;
			char name[FILE_MAX_NAME_LENGTH+1];
			char reserved[1];
		};
		char dataptr [4096];
	};

} FileInfo;

typedef struct __FileSystemInfo{
	union {
		struct {
			size_t diskSize;				// size of the disk
			size_t chunkSize;				// chunk Size
			size_t numDrives;				// number of drives
			size_t maxFiles;				// max num files
			size_t lastFileIndex;			// last opened file index
			size_t number_files_present;    // number of actual files contained by the fs
			FILEMODE dummy;					// space holder to keep FileInfo and FileSystemInfo same size
			TimeStamp creationDate;			// filesystem creation date
			char volname[FILE_MAX_NAME_LENGTH+1]; // volume name
			char reserved[1];
		};
		char dataptr [4096];
	};
} FileSystemInfo;
typedef union __FileSystemBlock{
		FileInfo        fileInfo [MAX_NUMBER_FILES+1];
		FileSystemInfo  filesystemInfo;
}FileSystemBlock;

typedef struct __FileSystem{
	int 				rawDeviceHandle;
	Disk*				diskInfo;
	boolean 			filesystemIsOpen;
	boolean 			filesystemInSync;
	void* 				IMFS_ptr;  				 // pointer
	size_t  			IMFS_Size; 				 // size in bytes needed to hold the filesystem info...
	boolean 			closeIsRegistered; // indicates whether a request to close on exit has been installed
	FileSystemBlock * 	fileSystemHeaderData;
} FileSystem;

typedef struct __MyFile{
	int    			index;
	FILEMODE 		mode;
	boolean 		isOpen;
	boolean 		reof;
	boolean 		lastRPacket;
	boolean 		lastWPacket;
	size_t  		lastRPacketRealSize;
	size_t  		lastWPacketRealSize;
	FileSystem * 	fs;
	struct aiocb 	writeOpInfo;
	struct aiocb 	readOpInfo;
} MyFile;

typedef struct __FileMetaData{
	char		tag[16];
	char		StartMPM[9];
	char		StartMJD[6];
	char		StopMPM[9];
	char		StopMJD[6];
	char		format[32];
	char		size[15];
	char		usage[15];
	char		complete[3];
	char		ignored;
} FileMetaData;

//#define fs (*((FileSystem*)fsbuf))
StatusCode FileSystem_Create(Disk * disk);
StatusCode FileSystem_Verify(FileSystem * fs);
StatusCode FileSystem_Open(FileSystem* fs, Disk * disk);
StatusCode FileSystem_Close(FileSystem * fs);
StatusCode FileSystem_ListFiles(FileSystem * fs);
size_t     FileSystem_GetFileCount(FileSystem * fs);
void       FileSystem_GetFileMetaData(FileSystem * fs, int index, FileMetaData* data);
void       FileSystem_SetFileMetaData(FileSystem * fs, int index, size_t startMPM, size_t startMJD, size_t reference, size_t StopMPM, size_t StopMJD, char * format);
boolean    FileSystem_IsFileCreatable(FileSystem * fs, char* name, size_t size, boolean overwrite);

StatusCode FileSystem_AsynchronousWriteStart(MyFile* file, char * srcPtr, size_t bytes);
StatusCode FileSystem_IsAsynchronousWriteDone(MyFile* file);
StatusCode FileSystem_GetAsynchronousBytesWritten(MyFile* file, size_t* result);
StatusCode FileSystem_AsynchronousReadStart(MyFile* file, char * dstPtr, size_t bytes);
StatusCode FileSystem_IsAsynchronousReadDone(MyFile* file);
StatusCode FileSystem_GetAsynchronousBytesRead(MyFile* file, size_t *result);

StatusCode FileSystem_CreateFile(FileSystem * fs, char* name, size_t size, boolean overwrite, int* fileIndex);
StatusCode FileSystem_OpenFile(FileSystem * fs, MyFile* file,char* fileName, FILEMODE mode);
StatusCode FileSystem_CloseFile(MyFile* file);
StatusCode FileSystem_DeleteFile(FileSystem * fs, int fileIndex);
boolean    FileSystem_FileExists(FileSystem * fs,char* fileName);

StatusCode _FileSystem_SynchronousWrite(FileSystem * fs, void* buf, off_t position, size_t count, ssize_t * result);
StatusCode _FileSystem_SynchronousRead(FileSystem * fs, void* buf, off_t position, size_t count, ssize_t * result);
StatusCode _FileSystem_AllocateIMFS(FileSystem * fs);
StatusCode _FileSystem_DeallocateIMFS(FileSystem * fs);

int        _FileSystem_GetFileIndex(FileSystem * fs, const char* name);
void 	   _FileSystem_RegisterActiveFilesystem(FileSystem * fs);
void 	   _FileSystem_UnregisterActiveFilesystem(FileSystem * fs);
void 	   _FiLeSystem_CloseAllActiveSystems();
boolean    _FileSystem_CheckPattern(void* ptr, size_t size);
void 	   _FileSystem_CreatePattern(void* ptr, size_t size);
size_t 	   _FileSystem_RangeConflicts(FileSystem * fs, size_t start, size_t stop);
boolean    _FileSystem_RegionOverlaps(size_t starta, size_t stopa, size_t startb, size_t stopb);
boolean    _FileSystem_IsBetween(size_t a, size_t b, size_t c);
void 	   _FileSystem_DumpFileInfo(FileSystem * fs, int index);
size_t FileSystem_GetFreeSpace(FileSystem * fs);

#endif /* FILESYSTEM_H_ */
