/*
 * Schedule.c
 *
 *  Created on: Oct 31, 2009
 *      Author: chwolfe2
 */
#include "Schedule.h"
#include "Time.h"
#include "Log.h"
#include "Persistence.h"
#include "Operations.h"
#include "Globals.h"
#include <assert.h>
#include <string.h>
#include <stdio.h>

Database scheduleDb;
boolean  isScheduleOpen=false;
boolean firstElementChanged=true;

void 	   Schedule_ConstructEntry(ScheduleEntry* entry, size_t reference, size_t startMJD, size_t startMPM, size_t stopMJD, size_t stopMPM, size_t datalength, char* data){
	assert(entry!=NULL);
	assert(data!=NULL);
	entry->data       = data;
	entry->reference  = reference;
	entry->startMJD   = startMJD;
	entry->startMPM   = startMPM;
	entry->stopMJD    = stopMJD;
	entry->stopMPM    = stopMPM;
	entry->datalength = datalength;
}


boolean _entryIsBeforeNow(ScheduleEntry* entry){
	assert(entry!=NULL);
	TimeStamp now=getTimestamp();
	size_t initMJD;
	size_t initMPM;
	// adjust for 2.5 seconds of init time
	if(entry->startMPM>2500){
		initMJD=entry->startMJD;
		initMPM=entry->startMPM-2500l;
	} else {
		initMJD=entry->startMJD-1l;
		initMPM=entry->startMPM+MILLISECONDS_PER_DAY-2500l;
	}
	return ((entry->startMJD < 5l) || before(initMJD, initMPM, now.MJD, now.MPM));
}
boolean _evBefore(ScheduleEntry* evA, ScheduleEntry* evB, size_t padding){
	assert(evA!=NULL && evB!=NULL);
	size_t aMJD;
	size_t aMPM;
	size_t bMJD;
	size_t bMPM;

	// adjust for 2.5 seconds of hold time
	if(evA->stopMPM+padding<MILLISECONDS_PER_DAY){
		aMJD=evA->stopMJD;			aMPM=evA->stopMPM + padding;
	} else {
		aMJD=evA->stopMJD+1l;		aMPM=evA->stopMPM + padding - MILLISECONDS_PER_DAY;
	}

	// adjust for 2.5 seconds of setup time
	if(evB->startMPM>padding){
		bMJD=evB->startMJD;			bMPM=evB->startMPM - padding;
	} else {
		bMJD=evB->startMJD-1l;		bMPM=evB->startMPM + MILLISECONDS_PER_DAY - padding;
	}
	return before(aMJD,aMPM,bMJD,bMPM);
}

boolean _evAfter(ScheduleEntry* evA, ScheduleEntry* evB, size_t padding){
	return _evBefore(evB,evA,padding);
}
boolean _evOverlaps(ScheduleEntry* evA, ScheduleEntry* evB, size_t padding){
	assert(evA!=NULL && evB!=NULL);
	size_t astartMJD;
	size_t astartMPM;
	size_t astopMJD;
	size_t astopMPM;
	size_t bstartMJD;
	size_t bstartMPM;
	size_t bstopMJD;
	size_t bstopMPM;

	// adjust for 2.5 seconds of setup time
	if(evA->startMPM>padding){
		astartMJD=evA->startMJD;		astartMPM=evA->startMPM - padding;
	} else {
		astartMJD=evA->startMJD-1l;		astartMPM=evA->startMPM + MILLISECONDS_PER_DAY - padding;
	}
	// adjust for 2.5 seconds of setup time
	if(evB->startMPM>padding){
		bstartMJD=evB->startMJD;		bstartMPM=evB->startMPM - padding;
	} else {
		bstartMJD=evB->startMJD-1l;		bstartMPM=evB->startMPM + MILLISECONDS_PER_DAY - padding;
	}

	// adjust for 2.5 seconds of hold time
	if(evA->stopMPM+padding<MILLISECONDS_PER_DAY){
		astopMJD=evA->stopMJD;			astopMPM=evA->stopMPM + padding;
	} else {
		astopMJD=evA->stopMJD+1l;		astopMPM=evA->stopMPM + padding - MILLISECONDS_PER_DAY;
	}
	// adjust for 2.5 seconds of hold time
	if(evB->stopMPM+padding<MILLISECONDS_PER_DAY){
		bstopMJD=evB->stopMJD;			bstopMPM=evB->stopMPM + padding;
	} else {
		bstopMJD=evB->stopMJD+1l;		bstopMPM=evB->stopMPM + padding - MILLISECONDS_PER_DAY;
	}
	if (before(astartMJD, astartMPM, bstopMJD, bstopMPM) && before(bstartMJD, bstartMPM, astartMJD, astartMPM)) return true;
	if (before(astopMJD,   astopMPM, bstopMJD, bstopMPM) && before(bstartMJD, bstartMPM,  astopMJD,  astopMPM)) return true;
	if (before(bstartMJD, bstartMPM, astopMJD, astopMPM) && before(astartMJD, astartMPM, bstartMJD, bstartMPM)) return true;
	if (before(bstopMJD,   bstopMPM, astopMJD, astopMPM) && before(astartMJD, astartMPM,  bstopMJD,  bstopMPM)) return true;
	if (equal(astartMJD, astartMPM, bstartMJD, bstartMPM)) return true;
	if (equal(astartMJD, astartMPM,  bstopMJD,  bstopMPM)) return true;
	if (equal(astopMJD,   astopMPM, bstartMJD, bstartMPM)) return true;
	if (equal(astopMJD,   astopMPM,  bstopMJD,  bstopMPM)) return true;
	return false;
}




int _Schedule_Count(){
	StatusCode sc;
	int count=0;
	sc = Persistence_List_GetCount(&scheduleDb, "SCHEDULEENTRY",&count);
	if (sc!=SUCCESS) count=-1;
	return count;
}


StatusCode _Schedule_Create(){
	StatusCode sc;
	sc = Persistence_Create(SCHEDULEFILENAME,NULL);					if (sc!=SUCCESS) return sc;
	sc = Persistence_Open(&scheduleDb,SCHEDULEFILENAME);			if (sc!=SUCCESS) return sc;
	sc = Persistence_List_Create(&scheduleDb, "SCHEDULEENTRY");		if (sc!=SUCCESS) return sc;
	sc = Persistence_List_Create(&scheduleDb, "SCHEDULEDATA");		if (sc!=SUCCESS) return sc;
	sc = Persistence_Close(&scheduleDb);
	return sc;
}
void _Schedule_Close(){
	if (isScheduleOpen){
		Persistence_Close(&scheduleDb);
		isScheduleOpen=false;
	}
}
StatusCode _Schedule_Open(){
	StatusCode sc = Persistence_Open(&scheduleDb,SCHEDULEFILENAME);
	if (sc!=SUCCESS){
		StatusCode sc = _Schedule_Create();
		if (sc!=SUCCESS){
			return FAILURE;
		} else {
			sc = Persistence_Open(&scheduleDb,SCHEDULEFILENAME);
			if (sc!=SUCCESS){
				return FAILURE;
			}
		}
	}


	int c =_Schedule_Count();
	if (c==-1) return FAILURE;
	// got to this point, then we're open
	ScheduleEntry entry;
	size_t bytes;
	boolean done=false;
	while (c>0 && !done) {
		sc = Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEENTRY",POSITION_FRONT,(char*)&entry,&bytes);
		if (sc!=SUCCESS || bytes!=sizeof(ScheduleEntry)) return FAILURE;
		if (_entryIsBeforeNow(&entry)){
			printf("Deleting expired entry: <ref=%lu, MJD=%lu, MPM=%lu>\n", entry.reference, entry.startMJD, entry.startMPM);
			Log_Add("Deleting expired entry: <ref=%lu, MJD=%lu, MPM=%lu>", entry.reference, entry.startMJD, entry.startMPM);
			sc = Persistence_List_RemoveItem(&scheduleDb, "SCHEDULEENTRY",POSITION_FRONT); if (sc!=SUCCESS) return FAILURE;
			sc = Persistence_List_RemoveItem(&scheduleDb, "SCHEDULEDATA",POSITION_FRONT);if (sc!=SUCCESS) return FAILURE;
			c =_Schedule_Count();			if (c==-1) return FAILURE;
		} else {
			done=true;
		}
	}
	atexit(_Schedule_Close);
	isScheduleOpen=true;
	return SUCCESS;
}

StatusCode Schedule_Open(boolean create){
	if (isScheduleOpen)return FAILURE;
	if (!create){
		return _Schedule_Open();
	} else {
		StatusCode sc = _Schedule_Create();	if (sc!=SUCCESS) return sc;
		return _Schedule_Open();
	}

}
void printentry(ScheduleEntry* entry){
	printf("Entry:\n"
			"\tStart MJD:%lu\n"
			"\tStart MPM:%lu\n"
			"\tStop  MJD:%lu\n"
			"\tStop  MPM:%lu\n"
			"\tDatalen:  %lu\n"
			"\tDataPtr:  %lu\n",
			entry->startMJD,
			entry->startMPM,
			entry->stopMJD,
			entry->stopMPM,
			entry->datalength,
			(size_t)entry->data);
}

StatusCode _Schedule_Insert(ScheduleEntry* entry, int position){
	StatusCode sc;
	assert(entry!=NULL);
	assert(entry->data!=NULL);
	sc = Persistence_List_AddItemBinary(&scheduleDb, "SCHEDULEENTRY",position,(char*)entry,sizeof(ScheduleEntry));
	if (sc!=SUCCESS) {
		printf("Scheduling Failure, insertion\n");
		Log_Add("Scheduling Failure, insertion");
		exit(EXIT_CODE_FAIL);
	}
	sc = Persistence_List_AddItemBinary(&scheduleDb, "SCHEDULEDATA",position,entry->data,entry->datalength);
	if (sc!=SUCCESS) {
		printf("Scheduling Failure, insertion\n");
		Log_Add("Scheduling Failure, insertion");
		exit(EXIT_CODE_FAIL);
	}
	if (position<=0 && position!=POSITION_BACK) firstElementChanged=true;
	return SUCCESS;
}



StatusCode Schedule_AddEntry(ScheduleEntry* entry, ScheduleEntry* conflict){
	//printf("AddEntry\n");
	//printf("entryptr %lu\n ",(size_t)entry);printentry(entry);
	assert(isScheduleOpen);
	StatusCode sc;
	int count = _Schedule_Count();
	if ((count==-1) || _entryIsBeforeNow(entry)) {
		memcpy((void *)conflict,(void *)entry,sizeof(ScheduleEntry));
		return FAILURE;
	}
	if (count==0) {
		//case where schedule is empty
		return  _Schedule_Insert(entry,POSITION_FRONT);
	}
	int i=0;
	size_t bytes=0;
	ScheduleEntry ToCompare1;
	ScheduleEntry ToCompare2;
	for(i=0;i<count;i++){
		sc = Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEENTRY",i,(char*)&ToCompare1,&bytes);
		if (sc!=SUCCESS || bytes!=sizeof(ScheduleEntry)) {
			printf("Scheduling Failure, entry missing\n");
			Log_Add("Scheduling Failure, entry missing");
			exit(EXIT_CODE_FAIL);
		}
		if (_evOverlaps(entry, &ToCompare1, 2500l)){
			memcpy((void *)conflict,(void *)&ToCompare1,sizeof(ScheduleEntry));
			return FAILURE;
		}
		//printf("AddEntry: i=%d : not overlaps\n",i);
		if (_evBefore(entry, &ToCompare1, 2500l)){
			// am before the next one in the list, so check to make sure we're after the previous item
			if (i==0){
				// case where before the first entry in the list
				return  _Schedule_Insert(entry,POSITION_FRONT);
			} else {
				// not the first, so check previous for after-ness
				sc = Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEENTRY",i-1,(char*)&ToCompare2,&bytes);
				if (sc!=SUCCESS || bytes!=sizeof(ScheduleEntry)) {
					printf("Scheduling Failure, entry missing\n");
					Log_Add("Scheduling Failure, entry missing");
					exit(EXIT_CODE_FAIL);
				}
				if (_evAfter(entry, &ToCompare1, 2500l)){
					// case where insertion point is in between i-1 and i
					return  _Schedule_Insert(entry,i);
				}
			}
		}
	}// end  of loop

	// if we got here, then entry is not before any entry, and doesn't conflict with any entry, thus it IS after all entrys, so add it
	// case where insertion point is in between i-1 and i
	return  _Schedule_Insert(entry,POSITION_BACK);
}
int Schedule_getIndexByReference(size_t reference){
	assert(isScheduleOpen);
	StatusCode sc;
	int count = _Schedule_Count();
	if (count==-1) return -1;
	int i=0;
	size_t bytes=0;
	ScheduleEntry temp;
	while (i<count){
		sc = Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEENTRY",i,(char*)&temp,&bytes);
		if (sc!=SUCCESS || bytes!=sizeof(ScheduleEntry)) {
			printf("Scheduling Failure, entry missing\n");
			Log_Add("Scheduling Failure, entry missing");
			exit(EXIT_CODE_FAIL);
		}
		if (temp.reference == reference) return i;
		i++;
	}
	return -1;
}
StatusCode Schedule_RemoveEntry(size_t reference){
	assert(isScheduleOpen);
	int idx=Schedule_getIndexByReference(reference);
	if (idx!=-1){
		StatusCode sc;
		sc = Persistence_List_RemoveItemBinary(&scheduleDb, "SCHEDULEDATA",idx);
		if (sc!=SUCCESS) {
			printf("Scheduling Failure, data missing\n");
			Log_Add("Scheduling Failure, data missing");
			exit(EXIT_CODE_FAIL);
		}
		sc = Persistence_List_RemoveItemBinary(&scheduleDb, "SCHEDULEENTRY",idx);
		if (sc!=SUCCESS) {
			printf("Scheduling Failure, entry missing\n");
			Log_Add("Scheduling Failure, entry missing");
			exit(EXIT_CODE_FAIL);
		}
		if (idx==0) firstElementChanged = true;
		return SUCCESS;
	} else {
		return FAILURE;
	}
}

StatusCode Schedule_GetEntry(size_t reference, ScheduleEntry* entry){
	assert(isScheduleOpen);
	assert(entry!=NULL);
	int idx=Schedule_getIndexByReference(reference);
	size_t bytes;
	if (idx!=-1){
		StatusCode sc;
		sc = Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEENTRY",idx,(char*)entry,&bytes);
		if (sc!=SUCCESS || bytes!=sizeof(ScheduleEntry)) {
			printf("Scheduling Failure, entry missing\n");
			Log_Add("Scheduling Failure, entry missing");
			exit(EXIT_CODE_FAIL);
		}
		return SUCCESS;
	} else {
		return FAILURE;
	}
}
StatusCode Schedule_GetEntryIndex(int index, ScheduleEntry* entry){
	assert(isScheduleOpen);
	assert(entry!=NULL);
	size_t bytes;
	return  Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEENTRY",index,(char*)entry,&bytes);
}
StatusCode Schedule_GetEntryDataIndex(int index, char* data, size_t* bytes){
	assert(isScheduleOpen);
	assert(data!=NULL);
	return  Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEDATA",index,data,bytes);
}

StatusCode Schedule_GetEntryData(size_t reference, char* data, size_t* bytes){
	assert(isScheduleOpen);
	assert(data!=NULL);
	int idx=Schedule_getIndexByReference(reference);
	if (idx!=-1){
		StatusCode sc;
		sc = Persistence_List_GetItemBinary(&scheduleDb, "SCHEDULEDATA",idx,data,bytes);
		if (sc!=SUCCESS) {
			printf("Scheduling Failure, data missing\n");
			Log_Add("Scheduling Failure, data missing");
			exit(EXIT_CODE_FAIL);
		}
		return SUCCESS;
	} else {
		return FAILURE;
	}
}

StatusCode Schedule_Close(){
	_Schedule_Close();
	return SUCCESS;
}

boolean Schedule_EntriesEqual(ScheduleEntry* a, ScheduleEntry* b){
	assert(a!=NULL && b!=NULL);
	if (a==b) return true;
	char * c=(char*)a;
	char * d=(char*)b;
	int i;
	for (i=0;i<sizeof(ScheduleEntry);i++){
		if (c[i]!=d[i]) return false;
	}
	return true;
}

boolean Schedule_Changed(){
	if (firstElementChanged){
		firstElementChanged=false;
		return true;
	}
	return false;
}




ScheduleEntry leadEntry;
OperationParameters leadEntryData;
boolean operationStarted=false;

#define SCHEDULER_STATE_IDLE       		0
#define SCHEDULER_STATE_WAIT_INIT  		1
#define SCHEDULER_STATE_WAIT_START 		2
#define SCHEDULER_STATE_WAIT_STOP 		3
#define SCHEDULER_STATE_WAIT_COOLDOWN	4
#define SCHEDULER_STATE_DELETE			5
#define SCHEDULER_STATE_ABORT			6
int schedulerState=SCHEDULER_STATE_IDLE;
void Process_AbortCurrent(){
	if (schedulerState!=SCHEDULER_STATE_IDLE){
		schedulerState=SCHEDULER_STATE_ABORT;
	}
}
void Process_Schedule(){
	//boolean tchanged=false;
	TimeStamp now = getTimestamp();
	/*static size_t lastMPM =0;
	static size_t lastMJD =0;
	if ((lastMPM!=now.MPM) || (lastMJD!=now.MJD)){
		lastMPM=now.MPM;
		lastMJD=now.MJD;
		tchanged=true;
	} else {
		tchanged=false;
	}*/

	size_t tempMPM = 0;
	size_t tempMJD = 0;

	switch (schedulerState){
		case SCHEDULER_STATE_IDLE       	:
			if (_Schedule_Count()>=1){
				if(Schedule_GetEntryIndex(POSITION_FRONT, &leadEntry)==SUCCESS){
					size_t bytes;
					if(Schedule_GetEntryDataIndex(POSITION_FRONT,(char*) &leadEntryData, &bytes)==SUCCESS || bytes!=sizeof(leadEntryData)){
						printf("[SCHEDULE] entered wait phase for operation with reference number %lu\n",leadEntryData.opReference );
						schedulerState=SCHEDULER_STATE_WAIT_INIT;
						Schedule_Changed(); // gotta check this so future checks will be valid.
					} else {
						printf("Error: Non-empty schedule and dequeue failed\n");
						exit(EXIT_CODE_FAIL);
					}
				} else {
					printf("Error: Non-empty schedule and dequeue failed\n");
					exit(EXIT_CODE_FAIL);
				}
			}
		break;
		case SCHEDULER_STATE_WAIT_INIT  	:
			if (Schedule_Changed()){
				printf("[SCHEDULE] schedule changed, abandoning operation with reference number %lu, return to idle\n",leadEntryData.opReference );
				schedulerState=SCHEDULER_STATE_IDLE; // go back to start and retrieve 1st scheduled entry
			} else {
				subTime(leadEntry.startMJD,leadEntry.startMPM,2500,&tempMJD,&tempMPM);
				if (!before(now.MJD,now.MPM,tempMJD,tempMPM)){
					//if (tchanged) printf("%lu:%lu >= %lu:%lu\n", now.MJD,now.MPM,tempMJD,tempMPM);
					printf("[SCHEDULE] entered init phase for operation with reference number %lu, waiting to init\n",leadEntryData.opReference );
					if(Operation_Init(&leadEntryData)==SUCCESS){
						printf("[SCHEDULE] initialized operation for operation with reference number %lu, waiting to start\n",leadEntryData.opReference );
						schedulerState=SCHEDULER_STATE_WAIT_START;
					} else {
						printf("[SCHEDULE]  initialization failure, aborting operation with reference number %lu\n",leadEntryData.opReference );
						schedulerState=SCHEDULER_STATE_ABORT;
					}
				} /*else {
					if (tchanged) printf("%lu:%lu < %lu:%lu\n", now.MJD,now.MPM,tempMJD,tempMPM);
				}*/
			}
		break;
		case SCHEDULER_STATE_WAIT_START 	:
			if (Schedule_Changed()){
				printf("[SCHEDULE] schedule changed, abandoning operation with reference number %lu, goto abort\n",leadEntryData.opReference );
				schedulerState=SCHEDULER_STATE_ABORT; // abort scheduled item
			} else {
				if (!before(now.MJD,now.MPM, leadEntry.startMJD,leadEntry.startMPM)){
					printf("[SCHEDULE] entered start phase for operation with reference number %lu\n",leadEntryData.opReference );
					schedulerState=SCHEDULER_STATE_WAIT_STOP;
				}
			}
		break;
		case SCHEDULER_STATE_WAIT_STOP 		:
			if (Schedule_Changed()){ //should only happen when lead entry is deleted!!!
				printf("[SCHEDULE] schedule changed, abandoning operation with reference number %lu, goto abort\n",leadEntryData.opReference );
				schedulerState=SCHEDULER_STATE_ABORT; // abort scheduled item
			} else {
				if (!before(now.MJD,now.MPM, globals.opData.opStopMJD, globals.opData.opStopMPM)){
					printf("---- opstop is %06lu:%09lu              ----now is %06lu:%09lu\n",globals.opData.opStopMJD, globals.opData.opStopMPM,now.MJD,now.MPM);
					printf("[SCHEDULE] entered stop phase for operation with reference number %lu\n",leadEntryData.opReference );
					if(Operation_CleanUp(&globals.opData)!=SUCCESS){
						printf("[SCHEDULE] operation cleanup failure\n");
						schedulerState=SCHEDULER_STATE_DELETE;
					} else {
						schedulerState=SCHEDULER_STATE_WAIT_COOLDOWN;
					}
				} else {
					if(Operation_Run(&globals.opData)!=SUCCESS){
						printf("[SCHEDULE] operation execution failure, aborting \n");
						schedulerState=SCHEDULER_STATE_ABORT;
					}
				}
			}
		break;
		case SCHEDULER_STATE_WAIT_COOLDOWN	:
			if (Schedule_Changed()){
				printf("[SCHEDULE] schedule changed, abandoning operation with reference number %lu, goto delete\n",leadEntryData.opReference );
				schedulerState=SCHEDULER_STATE_DELETE;
			} else {
				addTime(globals.opData.opStopMJD, globals.opData.opStopMPM,2500,&tempMJD,&tempMPM);
				if (!before(now.MJD,now.MPM, tempMJD, tempMPM)){
					printf("[SCHEDULE] cooldown ended for operation with reference number %lu, goto delete\n",leadEntryData.opReference );
					schedulerState=SCHEDULER_STATE_DELETE;
				}
			}
		break;
		case SCHEDULER_STATE_DELETE			:{
			printf("[SCHEDULE] deleting operation with reference number %lu from schedule, return to idle\n",leadEntryData.opReference );
			Schedule_RemoveEntry(leadEntry.reference);
			schedulerState=SCHEDULER_STATE_IDLE;
		}
		break;
		case SCHEDULER_STATE_ABORT			:
			printf("[SCHEDULE] E-stop operation with reference number %lu \n",leadEntryData.opReference );
			Operation_CleanUp(&globals.opData);
			printf("[SCHEDULE] E-delete operation with reference number %lu from schedule\n",leadEntryData.opReference );
			Schedule_RemoveEntry(leadEntry.reference);
			schedulerState=SCHEDULER_STATE_IDLE;
		break;

		default:
			printf("ERROR: Schedule FSM in undefined state\n");
			exit (EXIT_CODE_FAIL);
		break;
	}
}

